# MOFTransformer version 2.0.0
""" Vision Transformer (ViT) in PyTorch

A PyTorch implement of Vision Transformers as described in
'An Image Is Worth 16 x 16 Words: Transformers for Image Recognition at Scale' - https://arxiv.org/abs/2010.11929

The official jax code is released and available at https://github.com/google-research/vision_transformer

Acknowledgments:
* The paper authors for releasing code and weights, thanks!
* I fixed my class token impl based on Phil Wang's https://github.com/lucidrains/vit-pytorch ... check it out
for some einops/einsum fun
* Simple transformer style inspired by Andrej Karpathy's https://github.com/karpathy/minGPT
* Bert reference code checks against Huggingface Transformers and Tensorflow Bert

DeiT model defs and weights from https://github.com/facebookresearch/deit,
paper `DeiT: Data-efficient Image Transformers` - https://arxiv.org/abs/2012.12877

Hacked together by / Copyright 2020 Ross Wightman
"""
from functools import partial

import torch
import torch.nn as nn

from einops.layers.torch import Rearrange

from torch.nn import AvgPool3d
from timm.models.layers import DropPath, trunc_normal_


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        # NOTE scale factor was wrong in my original version, can set manually to be compat with prev weights
        self.scale = qk_scale or head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, mask=None):
        B, N, C = x.shape
        assert C % self.num_heads == 0
        qkv = (
            self.qkv(x)  # [B, N, 3*C]
            .reshape(
                B, N, 3, self.num_heads, C // self.num_heads
            )  # [B, N, 3, num_heads, C//num_heads]
            .permute(2, 0, 3, 1, 4)  # [3, B, num_heads, N, C//num_heads]
        )
        q, k, v = (
            qkv[0],  # [B, num_heads, N, C//num_heads]
            qkv[1],  # [B, num_heads, N, C//num_heads]
            qkv[2],  # [B, num_heads, N, C//num_heads]
        )  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale  # [B, num_heads, N, N]
        if mask is not None:
            mask = mask.bool()
            attn = attn.masked_fill(~mask[:, None, None, :], float("-inf"))
        attn = attn.softmax(dim=-1)  # [B, num_heads, N, N]
        attn = self.attn_drop(attn)

        x = (
            (attn @ v).transpose(1, 2).reshape(B, N, C)
        )  # [B, num_heads, N, C//num_heads] -> [B, N, C]
        x = self.proj(x)
        x = self.proj_drop(x)
        return x, attn


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x, mask=None):
        _x, attn = self.attn(self.norm1(x), mask=mask)
        x = x + self.drop_path(_x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x, attn


class PatchEmbed3D(nn.Module):
    """Image to Patch Embedding for 3D"""

    def __init__(
        self,
        img_size,  # minimum of H or W ex. 384
        patch_size,  # p -> length of fixed patch ex. 32
        in_chans=1,
        embed_dim=768,
        no_patch_embed_bias=False,
    ):
        super().__init__()

        assert img_size % patch_size == 0
        num_patches = (img_size**3) // (patch_size**3)
        self.img_size = img_size  # default: 30
        self.patch_size = patch_size  # default: 5
        self.num_patches = num_patches

        self.proj = nn.Sequential(
            Rearrange(
                "b c (h p1) (w p2) (d p3) -> b (h w d) (p1 p2 p3 c)",
                p1=patch_size,
                p2=patch_size,
                p3=patch_size,
            ),
            nn.Linear(patch_size * patch_size * patch_size * in_chans, embed_dim),
        )

    def forward(self, x):
        x = self.proj(x)  # [B, num_patches, embed_dim]
        return x


class VisionTransformer3D(nn.Module):
    """Vision Transformer

    A PyTorch impl of : `An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale`  -
        https://arxiv.org/abs/2010.11929
    """

    def __init__(
        self,
        img_size,
        patch_size,
        in_chans,
        embed_dim,
        depth=12,
        num_heads=12,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=None,
        add_norm_before_transformer=False,
        mpp_ratio=0.15,
        config=None,
    ):
        """
        Args:
            img_size (int, tuple): input image size
            patch_size (int, tuple): patch size
            in_chans (int): number of input channels
            embed_dim (int): embedding dimension
            depth (int): depth of transformer
            num_heads (int): number of attention heads
            mlp_ratio (int): ratio of mlp hidden dim to embedding dim
            qkv_bias (bool): enable bias for qkv if True
            qk_scale (float): override default qk scale of head_dim ** -0.5 if set
            drop_rate (float): dropout rate
            attn_drop_rate (float): attention dropout rate
            drop_path_rate (float): stochastic depth rate
            hybrid_backbone (nn.Module): CNN backbone to use in-place of PatchEmbed module
            norm_layer: (nn.Module): normalization layer
        """
        super().__init__()

        self.in_chans = in_chans
        self.mpp_ratio = mpp_ratio

        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        self.add_norm_before_transformer = add_norm_before_transformer

        self.patch_embed = PatchEmbed3D(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
        )
        num_patches = self.patch_embed.num_patches

        self.patch_size = patch_size  # default = 32
        self.patch_dim = img_size // patch_size
        self.mask_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)

        if add_norm_before_transformer:
            self.pre_norm = norm_layer(embed_dim)

        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, depth)
        ]  # stochastic depth decay rule
        self.blocks = nn.ModuleList(
            [
                Block(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[i],
                    norm_layer=norm_layer,
                )
                for i in range(depth)
            ]
        )
        self.norm = norm_layer(embed_dim)

        trunc_normal_(self.mask_token, std=0.02)
        trunc_normal_(self.pos_embed, std=0.02)
        trunc_normal_(self.cls_token, std=0.02)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def mask_tokens(self, orig_image, feats, patch_size, mpp_ratio):
        """
        Prepare masked tokens inputs/labels for masked patch prediction: 80% MASK, 10% random, 10% original.
        :param orig_image = _x, Tensor [B, C, H, W, D]
        :param feats = x  Tensor [B, ph*pw*pd, emb_dim]

        :return feats [B, ph*pw, emb_dim], labels [B, ph*pw, C]

        """

        m = AvgPool3d(patch_size, patch_size)
        with torch.no_grad():
            img_patch = m(orig_image)

        labels = (
            (img_patch.long().flatten(start_dim=2, end_dim=4))  # [B, C, ph*pw*pd]
            .permute(0, 2, 1)
            .contiguous()
        )  # [B, ph*pw*pd, C]

        # We sample a few tokens in each sequence for MLM training (with probability `self.mlm_probability`)
        # probability_matrix = torch.full(labels.shape[:-1], 0.15)  # [B, ph*pw*pd]
        probability_matrix = torch.full(labels.shape[:-1], mpp_ratio)  # [B, ph*pw*pd]
        masked_indices = torch.bernoulli(probability_matrix).bool()
        labels[
            ~masked_indices
        ] = -100  # We only compute loss on masked tokens [B, ph*pw*pd, C]
        """
        # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
        indices_replaced = (
                torch.bernoulli(torch.full(labels.shape[:-1], 0.8)).bool() & masked_indices
        )  # [B, ph*pw*pd]

        feats[indices_replaced] = self.mask_token.to(feats)
        """
        return feats, labels

    def visual_embed(self, _x, max_image_len, mask_it=False):
        """

        :param _x: batch images, Tensor [B, C, H, W, D]
        :param max_image_len: Int (or -1)
        :param mask_it: Bool
        :return:
            x:  Tensor [B, max_image_len+1, hid_dim],
            x_mask: Tensor [B, max_image_len+1]],
            (patch_index, (H, W, D)): [[B, max_image_len+1, 3], [H, W, D]]
            label: [B, max_image_len+1, C]
        """

        B, _, _, _, _ = _x.shape
        x = self.patch_embed(_x)  # [B, ph*pw*pd, embed_dim]
        # x = x.flatten(2).transpose(1, 2)

        # mpp
        if mask_it:
            x, label = self.mask_tokens(
                _x, x, self.patch_size, self.mpp_ratio
            )  # [B, ph*pw*pd, emb_dim], [B, ph*pw*pd, C]
            label = torch.cat(
                [torch.full((label.shape[0], 1, self.in_chans), -100).to(label), label],
                dim=1,
            )  # [B, max_len+1, C]

        # cls tokens
        cls_token = self.cls_token.expand(B, -1, -1)  # [B, 1, embed_dim]
        x = torch.cat([cls_token, x], dim=1)  # [B, ph*pw*pd, embed_dim]

        # positional embedding
        x += self.pos_embed
        x = self.pos_drop(x)

        if self.add_norm_before_transformer:
            x = self.pre_norm(x)

        x_mask = torch.ones(x.shape[:2]).to(x)  # [B, ph*pw*pd]

        if mask_it:
            return x, x_mask, label
        else:
            return x, x_mask, None



import pandas as pd
from pathlib import Path
from spbnet.utils.echo import title, param, start, end


def calcMeanStd(file_path: str, task: str, sample_num: int = 500):
    title("CALC MEAN STD")
    start("Star to calculate mean and std")
    file_path = Path(file_path)
    param(file=file_path.absolute(), task=task, sample_num=sample_num)
    df = pd.read_csv(file_path.absolute())
    df = df.dropna(subset=[task])
    df = df.sample(n=sample_num)
    mean = df[task].mean()
    std = df[task].std()
    end(f"Calculate end, mean={mean}, std={std}")
    title("CALC MEAN STD END")



import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import pearsonr
import seaborn as sns
from sklearn.metrics import r2_score
from einops import rearrange
import pandas as pd
from matplotlib.colors import LinearSegmentedColormap
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--dataset", type=str)
args = parser.parse_args()
dataset = args.dataset


def distance(pos1, pos2):
    def get_pos(a):
        az = a % 10
        a = a // 10
        ay = a % 10
        a = a // 10
        ax = a % 10
        a = a // 10
        return ax, ay, az

    ss = np.sum((np.array(get_pos(pos1)) - np.array(get_pos(pos2))) ** 2)
    dis = np.sqrt(ss)

    # if ss == 25 or ss == 50 or ss == 49:
    #     return dis
    # else:
    #     return 0

    return dis


show = np.load(f"./feat/{dataset}/self_attns.npy")
print("Load end, attn shape: ", show.shape)
# show = np.mean(show, axis=0)

# show = rearrange(show.reshape((10, 10, 10, -1)), "z y x a -> x y z a")
show = rearrange(show.reshape((-1, 10, 10, 10, 1000)), "b z y x a -> b x y z a")
indices = np.random.randint(0, show.shape[0], 1000)
show = show[indices]

# Heatmap

show = show.reshape((-1, 1000, 1000))
show = np.mean(show, axis=0)
# show = show[425:475, 425:475]
start, end = 400, 501
show = show[start:end, start:end]
# show = show[500]

cmap = LinearSegmentedColormap.from_list(
    "custom_cmap", list(zip([0, 1], ["#faf8fc", "#8387a1"]))
)

vmax = np.max(show) * 0.4
# plt.figure(figsize=(7, 6))
plt.imshow(show, cmap=cmap, aspect="auto", vmin=0, vmax=vmax)
plt.xlim([0, end - start])
plt.ylim([0, end - start])
plt.xticks(
    ticks=[i for i in range(0, end - start, 20)],
    labels=[i for i in range(start, end, 20)],
)
plt.yticks(
    ticks=[i for i in range(0, end - start, 20)],
    labels=[i for i in range(start, end, 20)],
)
# sns.kdeplot(
#     show,
#     cmap="mako",
#     fill=True,
#     thresh=0,
#     levels=100,
# )
plt.colorbar()
plt.title("Heatmap of self attention", fontsize=18, pad=20)
plt.ylabel("Patch of potential", fontsize=16, labelpad=10)
plt.xlabel("Patch of potential", fontsize=16, labelpad=10)
plt.savefig(
    "./png/self_attn.finegrained.png", dpi=300, format="png", bbox_inches="tight"
)
plt.savefig(
    "./eps/self_attn.finegrained.eps", dpi=600, format="eps", bbox_inches="tight"
)
# plt.savefig("./self_attn_center.png", dpi=300, format="png", bbox_inches="tight")
# plt.savefig("./self_attn.png", dpi=300, format="png", bbox_inches="tight")



from spbnet.data.calcMeanStd import calcMeanStd
import argparse


class CLICommand:
    """
    This command is used to estimate the mean and std of the dataset.

    There is no need for spbnet to use the command.
    SpbNet will automatically calculate the mean and variance from the training label data and normalize it.
    This command can be used to train baseline model PMTransformer or MOFTransforemr.

    To use this command, please prepare a csv file including label data.
    The csv file should look like:

    cifid   task1
    mof1    1.3
    mof2    2.1
    mof3    0.9
    ...     ...
    """

    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser):
        parser.add_argument("--file", type=str, help="path to csv file", required=True)
        parser.add_argument("--task", type=str, help="task name", required=True)
        parser.add_argument("--num", type=int, help="num to sample data", default=500)

    def run(args):
        calcMeanStd(args.file, args.task, args.num)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from transformers import (
    get_polynomial_decay_schedule_with_warmup,
    get_cosine_schedule_with_warmup,
    get_constant_schedule,
    get_constant_schedule_with_warmup,
)


def set_scheduler(pl_module):
    optim_config = pl_module.optimizer_config
    lr = optim_config["lr"]
    wd = optim_config["weight_decay"]

    no_decay = [
        "bias",
        "LayerNorm.bias",
        "LayerNorm.weight",
        "norm.bias",
        "norm.weight",
        "norm1.bias",
        "norm1.weight",
        "norm2.bias",
        "norm2.weight",
        "potential_mapper",
    ]
    head_names = ["head"]
    end_lr = optim_config["end_lr"]
    lr_mult = optim_config["lr_mult"]
    decay_power = optim_config["decay_power"]
    optim_type = optim_config["optim_type"]

    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if not any(nd in n for nd in no_decay)  # not within no_decay
                and not any(bb in n for bb in head_names)  # not within head_names
            ],
            "weight_decay": wd,
            "lr": lr,
        },
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if any(nd in n for nd in no_decay)  # within no_decay
                and not any(bb in n for bb in head_names)  # not within head_names
            ],
            "weight_decay": 0.0,
            "lr": lr,
        },
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if not any(nd in n for nd in no_decay)  # not within no_decay
                and any(bb in n for bb in head_names)  # within head_names
            ],
            "weight_decay": wd,
            "lr": lr * lr_mult,
        },
        {
            "params": [
                p
                for n, p in pl_module.named_parameters()
                if any(nd in n for nd in no_decay) and any(bb in n for bb in head_names)
                # within no_decay and head_names
            ],
            "weight_decay": 0.0,
            "lr": lr * lr_mult,
        },
    ]

    if optim_type == "adamw":
        optimizer = torch.optim.AdamW(
            optimizer_grouped_parameters, lr=lr, eps=1e-8, betas=(0.9, 0.98)
        )
    elif optim_type == "adam":
        optimizer = torch.optim.Adam(optimizer_grouped_parameters, lr=lr)
    elif optim_type == "sgd":
        optimizer = torch.optim.SGD(optimizer_grouped_parameters, lr=lr, momentum=0.9)

    if pl_module.trainer.max_steps == -1:
        max_steps = pl_module.trainer.estimated_stepping_batches
    else:
        max_steps = pl_module.trainer.max_steps

    warmup_steps = optim_config["warmup_steps"]
    if isinstance(optim_config["warmup_steps"], float):
        warmup_steps = int(max_steps * warmup_steps)

    print(
        f"max_epochs: {pl_module.trainer.max_epochs} | max_steps: {max_steps} | warmup_steps : {warmup_steps} "
        f"| weight_decay : {wd} | decay_power : {decay_power}"
    )

    if decay_power == "cosine":
        scheduler = get_cosine_schedule_with_warmup(
            optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=max_steps,
        )
    elif decay_power == "constant":
        scheduler = get_constant_schedule(
            optimizer,
        )
    elif decay_power == "constant_with_warmup":
        scheduler = get_constant_schedule_with_warmup(
            optimizer,
            num_warmup_steps=warmup_steps,
        )
    else:
        scheduler = get_polynomial_decay_schedule_with_warmup(
            optimizer,
            num_warmup_steps=warmup_steps,
            num_training_steps=max_steps,
            lr_end=end_lr,
            power=decay_power,
        )

    sched = {"scheduler": scheduler, "interval": "step"}

    return (
        [optimizer],
        [sched],
    )


# def get_scheduler(pl_module, optimizer):
#     if pl_module.trainer.max_steps == -1:
#         max_steps = pl_module.trainer.estimated_stepping_batches
#     else:
#         max_steps = pl_module.trainer.max_steps
#     warmup_steps = pl_module.optimizer_config["warmup_steps"]
#     if isinstance(pl_module.optimizer_config["warmup_steps"], float):
#         warmup_steps = int(max_steps * warmup_steps)

#     decay_power = pl_module.optimizer_config["decay_power"]
#     end_lr = pl_module.optimizer_config["end_lr"]

#     if decay_power == "cosine":
#         scheduler = get_cosine_schedule_with_warmup(
#             optimizer,
#             num_warmup_steps=warmup_steps,
#             num_training_steps=max_steps,
#         )
#     elif decay_power == "constant":
#         scheduler = get_constant_schedule(
#             optimizer,
#         )
#     elif decay_power == "constant_with_warmup":
#         scheduler = get_constant_schedule_with_warmup(
#             optimizer,
#             num_warmup_steps=warmup_steps,
#         )
#     else:
#         scheduler = get_polynomial_decay_schedule_with_warmup(
#             optimizer,
#             num_warmup_steps=warmup_steps,
#             num_training_steps=max_steps,
#             lr_end=end_lr,
#             power=decay_power,
#         )

#     return scheduler



# MOFTransformer version 2.0.0
import re
from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()


with open("requirements.txt", "r") as f:
    install_requires = f.readlines()
# extras_require = {"docs": ["sphinx", "livereload", "myst-parser"]}

with open("spbnet/__init__.py") as f:
    version = re.search(r"__version__ = [\'\"](?P<version>.+)[\.\"]", f.read()).group(
        "version"
    )


setup(
    name="spbnet",
    version=version,
    description="spbnet",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jiawen Zou",
    author_email="20307130189@fudan.edu.cn",
    packages=find_packages(),
    package_data={
        "spbnet": [
            "data/libs/**/*",
            "**/*.yaml",
            "visualize/template/**.html",
            # "libs/GRIDAY/*",
            # "libs/GRIDAY/scripts/*",
            # "libs/GRIDAY/FF/*",
            # "assets/*.json",
            # "examples/dataset/*",
            # "examples/dataset/**/*",
            # "examples/raw/*",
            # "examples/visualize/dataset/*",
            # "examples/visualize/dataset/test/*",
        ]
    },
    install_requires=install_requires,
    # extras_require=extras_require,
    scripts=[],
    url="https://tyvanzou.github.io/spbnet/",
    download_url="https://github.com/tyvanzou/spbnet",
    entry_points={
        "console_scripts": [
            "spbnet=spbnet.cli.main:main",
        ]
    },
    # python_requires=">=3.8",
)



import os
import math
import logging
import logging.handlers
import subprocess
import pickle
from pathlib import Path

import numpy as np

from pymatgen.io.cif import CifParser
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.cssr import Cssr

from ase.io import read
from ase.neighborlist import natural_cutoffs
from ase import neighborlist
from ase.build import make_supercell


cur_dir = Path(__file__).parent

GRIDAY_PATH = os.path.join(cur_dir, "libs/GRIDAYS/scripts/grid_gen")
FF_PATH = os.path.join(cur_dir, "libs/GRIDAYS/FF")


def get_logger(filename):
    logger = logging.getLogger(filename)
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter(
        fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    return logger


def get_unique_atoms(atoms):
    # get graph
    cutoff = natural_cutoffs(atoms)
    neighbor_list = neighborlist.NeighborList(
        cutoff, self_interaction=True, bothways=True
    )
    neighbor_list.update(atoms)
    matrix = neighbor_list.get_connectivity_matrix()

    # Get N, N^2
    numbers = atoms.numbers
    number_sqr = np.multiply(numbers, numbers)

    matrix_sqr = matrix.dot(matrix)
    matrix_cub = matrix_sqr.dot(matrix)
    matrix_sqr.data[:] = 1  # count 1 for atoms

    # calculate
    list_n = [numbers, number_sqr]
    list_m = [matrix, matrix_sqr, matrix_cub]

    arr = [numbers]

    for m in list_m:
        for n in list_n:
            arr.append(m.dot(n))

    arr = np.vstack(arr).transpose()

    uni, unique_idx, unique_count = np.unique(
        arr, axis=0, return_index=True, return_counts=True
    )

    # sort
    final_uni = uni[np.argsort(-unique_count)].tolist()
    final_unique_count = unique_count[np.argsort(-unique_count)].tolist()

    arr = arr.tolist()
    final_unique_idx = []
    for u in final_uni:
        final_unique_idx.append([i for i, a in enumerate(arr) if a == u])

    return final_unique_idx, final_unique_count


def get_crystal_graph(atoms, radius=8, max_num_nbr=12):
    dist_mat = atoms.get_all_distances(mic=True)
    nbr_mat = np.where(dist_mat > 0, dist_mat, 1000)  # 1000 is mamium number
    nbr_idx = []
    nbr_dist = []
    for row in nbr_mat:
        idx = np.argsort(row)[:max_num_nbr]
        nbr_idx.extend(idx)
        nbr_dist.extend(row[idx])

    # get same-topo atoms
    uni_idx, uni_count = get_unique_atoms(atoms)

    # convert to small size
    atom_num = np.array(list(atoms.numbers), dtype=np.int8)
    nbr_idx = np.array(nbr_idx, dtype=np.int16)
    nbr_dist = np.array(nbr_dist, dtype=np.float32)
    uni_count = np.array(uni_count, dtype=np.int16)
    return atom_num, nbr_idx, nbr_dist, uni_idx, uni_count


def _calculate_scaling_matrix_for_orthogonal_supercell(cell_matrix, eps=0.01):
    """
    cell_matrix: contains lattice vector as column vectors.
                 e.g. cell_matrix[:, 0] = a.
    eps: when value < eps, the value is assumed as zero.
    """
    inv = np.linalg.inv(cell_matrix)

    # Get minimum absolute values of each row.
    abs_inv = np.abs(inv)
    mat = np.where(abs_inv < eps, np.full_like(abs_inv, 1e30), abs_inv)
    min_values = np.min(mat, axis=1)

    # Normalize each row with minimum absolute value of each row.
    normed_inv = inv / min_values[:, np.newaxis]

    # Calculate scaling_matrix.
    # New cell = np.dot(scaling_matrix, cell_matrix).
    scaling_matrix = np.around(normed_inv).astype(np.int32)

    return scaling_matrix


def get_energy_grid(atoms, cif_id, root_dataset, eg_logger):
    # Before 1.1.1 version : num_grid = [str(round(cell)) for cell in structure.lattice.abc]
    # After 1.1.1 version : num_grid = [30, 30, 30]
    global GRIDAY_PATH, FF_PATH

    eg_file = os.path.join(root_dataset, cif_id)
    # random_str = str(np.random.rand()).encode()
    tmp_file = os.path.join(root_dataset, f"{cif_id}.cssr")

    try:
        structure = AseAtomsAdaptor().get_structure(atoms)
        Cssr(structure).write_file(tmp_file)
        if not os.path.exists(tmp_file):
            eg_logger.info(f"{cif_id} cssr write fail")
            return False
        num_grid = ["30", "30", "30"]
        proc = subprocess.Popen(
            [
                GRIDAY_PATH,
                *num_grid,
                f"{FF_PATH}/UFF_Type.def",
                f"{FF_PATH}/UFF_FF.def",
                tmp_file,
                eg_file,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = proc.communicate()
    finally:
        # remove temp_file
        if os.path.exists(tmp_file):
            os.remove(tmp_file)

    if err:
        eg_logger.info(f"{cif_id} energy grid failed {err}")
        return False
    else:
        eg_logger.info(f"{cif_id} energy grid success")

    return True


def _make_supercell(atoms, cutoff):
    """
    make atoms into supercell when cell length is less than cufoff (min_length)
    """
    # when the cell lengths are smaller than radius, make supercell to be longer than the radius
    scale_abc = []
    for l in atoms.cell.cellpar()[:3]:
        if l < cutoff:
            scale_abc.append(math.ceil(cutoff / l))
        else:
            scale_abc.append(1)

    # make supercell
    m = np.zeros([3, 3])
    np.fill_diagonal(m, scale_abc)
    atoms = make_supercell(atoms, m)
    return atoms


def make_prepared_data(
    cif: Path, root_dataset_total: Path, logger=None, eg_logger=None, **kwargs
):
    if logger is None:
        logger = get_logger(filename="prepare_data.log")
    if eg_logger is None:
        eg_logger = get_logger(filename="prepare_energy_grid.log")

    if isinstance(cif, str):
        cif = Path(cif)
    if isinstance(root_dataset_total, str):
        root_dataset_total = Path(root_dataset_total)

    root_dataset_total.mkdir(exist_ok=True, parents=True)

    max_length = kwargs.get("max_length", 60.0)
    min_length = kwargs.get("min_length", 30.0)
    max_num_nbr = kwargs.get("max_num_nbr", 12)
    max_num_unique_atoms = kwargs.get("max_num_unique_atoms", 300)
    max_num_atoms = kwargs.get("max_num_atoms", None)

    cif_id: str = cif.stem

    p_graphdata = root_dataset_total / f"{cif_id}.graphdata"
    p_griddata = root_dataset_total / f"{cif_id}.griddata16"
    p_grid = root_dataset_total / f"{cif_id}.grid"

    # # Grid data and Graph data already exists
    # if p_graphdata.exists() and p_griddata.exists() and p_grid.exists():
    #     logger.info(f"{cif_id} graph data already exists")
    #     eg_logger.info(f"{cif_id} energy grid already exists")
    #     return True

    # valid cif check
    try:
        CifParser(cif).get_structures()
    except ValueError as e:
        logger.info(f"{cif_id} failed : {e} (error when reading cif with pymatgen)")
        return False

    # read cif by ASE
    try:
        atoms = read(str(cif))
    except Exception as e:
        logger.error(f"{cif_id} failed : {e}")
        return False

    # 1. get crystal graph
    atoms = _make_supercell(atoms, cutoff=8)  # radius = 8

    if max_num_atoms and len(atoms) > max_num_atoms:
        logger.error(
            f"{cif_id} failed : number of atoms are larger than `max_num_atoms` ({max_num_atoms})"
        )
        return False

    atom_num, nbr_idx, nbr_dist, uni_idx, uni_count = get_crystal_graph(
        atoms, radius=8, max_num_nbr=max_num_nbr
    )
    if len(nbr_idx) < len(atom_num) * max_num_nbr:
        logger.error(
            f"{cif_id} failed : num_nbr is smaller than max_num_nbr. please make radius larger"
        )
        return False

    if len(uni_idx) > max_num_unique_atoms:
        logger.error(
            f"{cif_id} failed : The number of topologically unique atoms is larget than `max_num_unique_atoms` ({max_num_unique_atoms})"
        )
        return False

    # 2. make supercell with min_length
    atoms_eg = _make_supercell(atoms, cutoff=min_length)
    for l in atoms_eg.cell.cellpar()[:3]:
        if l > max_length:
            logger.error(f"{cif_id} failed : supercell have more than max_length")
            return False

    # 3. calculate energy grid
    eg_success = get_energy_grid(atoms_eg, cif_id, root_dataset_total, eg_logger)

    if eg_success:
        logger.info(f"{cif_id} succeed : supercell length {atoms.cell.cellpar()[:3]}")

        # save cif files
        save_cif_path = root_dataset_total / f"{cif_id}.cif"
        atoms.write(filename=save_cif_path)

        # save graphdata file
        data = [cif_id, atom_num, nbr_idx, nbr_dist, uni_idx, uni_count]
        with open(str(p_graphdata), "wb") as f:
            pickle.dump(data, f)
        return True
    else:
        return False



from typing import Any, Optional
from functools import partial

# from pytorch_lightning.utilities.types import STEP_OUTPUT
# from model.crossformer import CrossFormer
from spbnet.modules.module import CrossFormer
from spbnet.modules.heads import RegressionHead, ClassificationHead, Pooler
from spbnet.modules.optimize import set_scheduler
from spbnet.modules import objectives
from torch import optim, nn
from data.dataset_pretrain import Dataset
import pytorch_lightning as pl
import pandas as pd
import yaml
import torch
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.callbacks import LearningRateMonitor
from pytorch_lightning.loggers import TensorBoardLogger
from torch.utils.data import DataLoader
from einops import rearrange
from sklearn.model_selection import train_test_split
from pathlib import Path
from spbnet.utils.echo import err, title, start, end, param


cur_dir = Path(__file__).parent


class CrossFormerTrainer(pl.LightningModule):
    def __init__(self, config: dict):
        super(CrossFormerTrainer, self).__init__()
        self.save_hyperparameters()

        self.config = config
        self.model_config = config
        self.optimizer_config = config

        model_config = self.model_config
        self.model = CrossFormer(model_config)
        # print(self.model)

        # pooler
        self.pooler = Pooler(model_config["hid_dim"])
        self.pooler.apply(objectives.init_weights)

        # void fraction prediction
        self.vfp_head = RegressionHead(model_config["hid_dim"])
        self.vfp_head.apply(objectives.init_weights)

        # topo classify
        self.tc_head = ClassificationHead(model_config["hid_dim"], config["topo_num"])
        self.tc_head.apply(objectives.init_weights)

        # atom grid classify
        self.agc_head = nn.Linear(model_config["hid_dim"], 1)
        self.agc_head.apply(objectives.init_weights)

        self.mse_loss = nn.MSELoss()
        self.mae_loss = nn.L1Loss()
        self.cross_entropy = nn.CrossEntropyLoss()

    def training_step(self, batch, batch_idx) -> torch.Tensor:
        # lr log
        lrs = self.lr_schedulers().get_last_lr()
        lr_avg = sum(lrs)
        lr_avg /= len(lrs)
        self.log("lr", lr_avg)

        # [batch_size, max_token_len]
        vf = batch["voidfraction"]  # [B] float
        topo = batch["topo"]  # [B] int
        atomgrid = batch["atomgrid"]  # [B, GRID, GRID, GRID] aka [B, 30, 30, 30]

        feat = self.model(batch)

        cls_feat = feat["cls_feat"]
        cls_feat = self.pooler(cls_feat)

        # vfp
        vfp = self.vfp_head(cls_feat)  # [batch_size, 1]
        vfp = vfp.reshape(vfp.shape[0])  # [batch_size]
        vf = vf.reshape(vf.shape[0])  # [batch_size]
        vfp_mse = self.mse_loss(vfp, vf)
        vfp_mae = self.mae_loss(vfp, vf)
        self.log("train_vfp_mse", vfp_mse, batch_size=vf.shape[0], sync_dist=True)
        self.log("train_vfp_mae", vfp_mae, batch_size=vf.shape[0], sync_dist=True)

        # tc
        topo_pred = self.tc_head(cls_feat)  # [batch_size, topo_num]
        topo = topo.reshape(vf.shape[0])  # [batch_size]
        tc_loss = self.cross_entropy(topo_pred, topo)
        # tc_loss = self.focal_loss(topo_pred, topo)
        tc_acc = self.cal_acc(topo, topo_pred)
        self.log("train_tc_loss", tc_loss, batch_size=topo.shape[0], sync_dist=True)
        self.log("train_tc_acc", tc_acc, batch_size=topo.shape[0], sync_dist=True)

        # atom grid
        potential_feat = feat[
            "potential_feat"
        ]  # [B, GRID / PathSize, GRID / PathSize, GRID / PathSize, hid_dim] aka [B, 10, 10, 10, 768]
        agc_pred = self.agc_head(potential_feat)
        agc_label: torch.Tensor = atomgrid
        patch_size = self.model_config["patch_size"]["lj"]
        # NOTE: Correct
        # agc_label = agc_label.transpose(-1, -3) # [b x y z] -> [b h w d]
        agc_label = rearrange(
            agc_label,
            "b (h p1) (w p2) (d p3) -> b (h w d) (p1 p2 p3)",
            p1=patch_size,
            p2=patch_size,
            p3=patch_size,
        )
        agc_label = agc_label.sum(
            -1, keepdim=True
        )  # [B, GRID / PatchSize, GRID / PatchSize, GRID / PatchSize, 1]
        agc_mse = self.mse_loss(agc_label, agc_pred)
        agc_mae = self.mae_loss(agc_label, agc_pred)
        self.log(
            "train_agc_mse", agc_mse, batch_size=agc_label.shape[0], sync_dist=True
        )
        self.log(
            "train_agc_mae", agc_mae, batch_size=agc_label.shape[0], sync_dist=True
        )

        loss = (
            self.optimizer_config["vfp"] * vfp_mse
            + self.optimizer_config["tc"] * tc_loss
            + self.optimizer_config["agc"] * agc_mse
        )
        return loss

    def validation_step(self, batch, batch_idx) -> torch.Tensor:
        # [batch_size, max_token_len]
        cifid = batch["cifid"]
        vf = batch["voidfraction"]  # [B] float
        topo = batch["topo"]  # [B] int
        atomgrid = batch["atomgrid"]  # [B, GRID, GRID, GRID] aka [B, 30, 30, 30]

        feat = self.model(batch)

        cls_feat = feat["cls_feat"]
        cls_feat = self.pooler(cls_feat)

        # vfp
        vfp = self.vfp_head(cls_feat)  # [batch_size, 1]
        vfp = vfp.reshape(vfp.shape[0])  # [batch_size]
        vf = vf.reshape(vf.shape[0])  # [batch_size]
        vfp_mse = self.mse_loss(vfp, vf)
        vfp_mae = self.mae_loss(vfp, vf)
        self.log("val_vfp_mse", vfp_mse, batch_size=vf.shape[0], sync_dist=True)
        self.log("val_vfp_mae", vfp_mae, batch_size=vf.shape[0], sync_dist=True)

        # tc
        topo_pred = self.tc_head(cls_feat)  # [batch_size, topo_num]
        topo = topo.reshape(vf.shape[0])  # [batch_size]
        tc_loss = self.cross_entropy(topo_pred, topo)
        # tc_loss = self.focal_loss(topo_pred, topo)
        tc_acc = self.cal_acc(topo, topo_pred)
        self.log("val_tc_loss", tc_loss, batch_size=topo.shape[0], sync_dist=True)
        self.log("val_tc_acc", tc_acc, batch_size=topo.shape[0], sync_dist=True)

        # atom grid classify
        potential_feat = feat[
            "potential_feat"
        ]  # [B, GRID / PathSize, GRID / PathSize, GRID / PathSize, hid_dim] aka [B, 10, 10, 10, 768]
        agc_pred = self.agc_head(potential_feat)
        agc_label: torch.Tensor = atomgrid
        patch_size = self.model_config["patch_size"]["lj"]
        agc_label = rearrange(
            agc_label,
            "b (h p1) (w p2) (d p3) -> b (h w d) (p1 p2 p3)",
            p1=patch_size,
            p2=patch_size,
            p3=patch_size,
        )
        agc_label = agc_label.sum(
            -1, keepdim=True
        )  # [B, GRID / PatchSize, GRID / PatchSize, GRID / PatchSize, 1]
        agc_mse = self.mse_loss(agc_label, agc_pred)
        agc_mae = self.mae_loss(agc_label, agc_pred)
        self.log("val_agc_mse", agc_mse, batch_size=agc_label.shape[0], sync_dist=True)
        self.log("val_agc_mae", agc_mae, batch_size=agc_label.shape[0], sync_dist=True)

    def cal_acc(self, y, pred):
        predicted_labels = torch.argmax(pred, dim=-1)
        correct_predictions = (predicted_labels == y).int()
        accuracy = correct_predictions.float().mean()
        return accuracy

    def configure_optimizers(self) -> Any:
        return set_scheduler(self)


def pretrain(config_path: Path):
    if type(config_path) is str:
        config_path = Path(config_path)
    if not config_path.exists():
        err(f"config_path: {config_path.absolute()} not exists!")

    with open((cur_dir / "config.pretrain.yaml").absolute(), "r") as f:
        base_config = yaml.load(f, Loader=yaml.FullLoader)

    with open(config_path.absolute(), "r") as f:
        user_config = yaml.load(f, Loader=yaml.FullLoader)

    if user_config.get("data_dir") is None:
        err(f"Please specify modal directory `data_dir`!")
        return
    if user_config.get("id_prop") is None:
        err(f"Please specify label data `id_prop`")
        return
    if user_config.get("log_dir") is None:
        err(f"Please specify log directory")
        return

    base_config.update(user_config)
    config = base_config
    param(**config)

    title("START TO PRETRAIN")

    df = pd.read_csv(config["id_prop"])
    train_df, val_df = train_test_split(df, test_size=0.1, random_state=42)
    data_dir = Path(config["data_dir"])

    train_dataset = Dataset(train_df, data_dir, config["nbr_fea_len"])
    val_dataset = Dataset(val_df, data_dir, config["nbr_fea_len"])

    config["topo_num"] = train_dataset.get_topo_num()

    checkpoint_callback = ModelCheckpoint(
        monitor="val_agc_mae", mode="min", save_last=True
    )
    logger = TensorBoardLogger(save_dir=config["log_dir"], name="")
    lr_monitor = LearningRateMonitor(logging_interval="epoch")
    trainer = pl.Trainer(
        max_epochs=config["epoch"],
        min_epochs=0,
        devices=config["cuda"],
        accelerator=config["accelerator"],
        strategy=config["strategy"],
        callbacks=[checkpoint_callback, lr_monitor],
        accumulate_grad_batches=config["accumulate_grad_batches"],
        precision=config["precision"],
        log_every_n_steps=config["log_every_n_steps"],
        logger=logger,
    )
    model = CrossFormerTrainer(config)
    # ckpt = torch.load(
    #     "./lightning_logs/scratch/co2_2.5bar/shaper/checkpoints/epoch=146-step=13818.ckpt"
    # )
    # model.load_state_dict(ckpt["state_dict"], strict=False)
    trainer.fit(
        model=model,
        train_dataloaders=DataLoader(
            train_dataset,
            batch_size=config["batch_size"],
            shuffle=True,
            num_workers=8,
            collate_fn=partial(Dataset.collate, img_size=config["img_size"]),
            persistent_workers=True,
        ),
        val_dataloaders=DataLoader(
            val_dataset,
            batch_size=config["batch_size"],
            shuffle=False,
            num_workers=8,
            collate_fn=partial(Dataset.collate, img_size=config["img_size"]),
            persistent_workers=True,
        ),
        # ckpt_path="./lightning_logs/version_2/checkpoints/last.ckpt",
    )



# MOFTransformer version 2.1.0
import torch
import torch.nn as nn
import torch.nn.functional as F


def init_weights(module):
    if isinstance(module, (nn.Linear, nn.Embedding)):
        module.weight.data.normal_(mean=0.0, std=0.02)
    elif isinstance(module, nn.LayerNorm):
        module.bias.data.zero_()
        module.weight.data.fill_(1.0)

    if isinstance(module, nn.Linear) and module.bias is not None:
        module.bias.data.zero_()


from spbnet.data.buildData import buildData
import argparse


class CLICommand:
    """
    Build datas of different modals, since spbnet is a multi-modal network.

    Please provide directory like:

    - root_dir
        - cif
            - mof1.cif
            - mof2.cif
            ...

    This command will produce:

    - root_dir
        - cif
            - mof1.cif
            - mof2.cif
            ...
        - spbnet
            - graphdata
                - mof1.graphdata
                - mof2.graphdata
                ...
            - grid
                - mof1.grid
                - mof2.grid
                ...
            - griddata8
                - mof1.npy
                - mof2.npy
                ...
    """

    @staticmethod
    def add_arguments(parser: argparse.ArgumentParser):
        parser.add_argument(
            "--root-dir", type=str, help="root_dir to build modal data", required=True
        )
        parser.add_argument(
            "--cif-dir",
            type=str,
            help='cif_dir to use, default is "cif", this command will use cif files from root_dir/cif_dir',
            default="cif",
        )
        parser.add_argument(
            "--target-dir",
            type=str,
            help="target directory to build data, this command will produce data in root_dir/target_dir",
            default="spbnet",
        )
        parser.add_argument(
            "--n-process",
            type=int,
            help="multiprocessing, number of processes to use, default is 1",
            default=1,
        )
        parser.add_argument(
            "--crystal-max-length",
            type=float,
            help="In order to ensure that the energy grid resolution is greater than 3A, a supercell needs to be constructed. crystal_max_length is the maximum lattice constant length of the supercell, the default is 120.0. See papers of SpbNet and MOFTransformer for more information",
            default=120.0,
        )
        parser.add_argument(
            "--max-num-unique-atoms",
            help="A crystal with too many atoms in a single unit cell will lead to large errors, so the number of atoms needs to be limited. The default limit is 512. For tobacco dataset we suggest to use 1024. See MOFTransformer paper for more information",
            type=int,
            default=512,
        )

    @staticmethod
    def run(args):
        buildData(
            args.root_dir,
            args.cif_dir,
            args.target_dir,
            args.n_process,
            args.crystal_max_length,
            args.max_num_unique_atoms,
        )
